/* Detokenize MAC/65 Files
 * Copyright (C) 2010 by Ivo van Poorten <ivop@free.fr>, 2010-01-21
 * General Public License version 2
 * See: http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

static char *myname;

void fatal(char *fmt, ...) {
    va_list ap;
    fflush(stdout);
    va_start(ap, fmt);
    fprintf(stderr, "%s: ", myname);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(1);
}

#define GET16(f)    (fgetc(f)+(fgetc(f)<<8))

static const char * const tok1[256] = {
    "ERROR -",  ".IF",      ".ELSE",    ".ENDIF",   /*  0... */
    ".MACRO",   ".ENDM",    ".TITLE",   "",
    ".PAGE",    ".WORD",    ".ERROR",   ".BYTE",
    ".SBYTE",   ".DBYTE",   ".END",     ".OPT",
    ".TAB",     ".INCLUDE", ".DS",      ".ORG",     /* 16... */
    ".EQU",     "BRA",      "TRB",      "TSB",
    ".FLOAT",   ".CBYTE",   ";",        ".LOCAL",
    ".SET",     "*=",       "=",        ".=",
    "JSR",      "JMP",      "DEC",      "INC",      /* 32... */
    "LDX",      "LDY",      "STX",      "STY",
    "CPX",      "CPY",      "BIT",      "BRK",
    "CLC",      "CLD",      "CLI",      "CLV",
    "DEX",      "DEY",      "INX",      "INY",      /* 48... */
    "NOP",      "PHA",      "PHP",      "PLA",
    "PLP",      "RTI",      "RTS",      "SEC",
    "SED",      "SEI",      "TAX",      "TAY",
    "TSX",      "TXA",      "TXS",      "TYA",      /* 64... */
    "BCC",      "BCS",      "BEQ",      "BMI",
    "BNE",      "BPL",      "BVC",      "BVS",
    "ORA",      "AND",      "EOR",      "ADC",
    "STA",      "LDA",      "CMP",      "SBC",      /* 80... */
    "ASL",      "ROL",      "LSR",      "ROR",
    "",         "STZ",      "DEA",      "INA",
    "PHX",      "PHY",      "PLX",      "PLY"
};

static const char * const tok2[256] = {
    [ 5] = "",      [ 6] = "",      [ 7] = "",      [ 8] = "",      [10] = "",
    [11] = "%%$",   [12] = "%%",    [13] = "*",     [18] = "+",
    [19] = "-",     [20] = "*",     [21] = "/",     [22] = "&",
    [24] = "=",     [25] = "<=",    [26] = ">=",    [27] = "<>",
    [28] = ">",     [29] = "<",     [30] = "-",     [31] = "[",
    [32] = "]",     [36] = "!",     [37] = "^",     [39] = "\\",
    [47] = ".REF",  [48] = ".DEF",  [49] = ".NOT",  [50] = ".AND",
    [51] = ".OR",   [52] = "<",     [53] = ">",     [54] = ",X)",
    [55] = "),Y",   [56] = ",Y",    [57] = ",X",    [58] = ")",
    [59] = "",      [61] = ",",     [62] = "#",     [63] = "A",
    [64] = "(",     [65] = "\"",    [69] = "NO",    [70] = "OBJ",
    [71] = "ERR",   [72] = "EJECT", [73] = "LIST",  [74] = "XREF",
    [75] = "MLIST", [76] = "CLIST", [77] = "NUM"
};

int main(int argc, char **argv) {
    unsigned char line[256];
    FILE *f;
    long flen;
    int lnum, len, x, y;

    myname = argv[0];

    if (argc !=2)
        fatal("usage: %s filename\n", myname);

    if (!(f = fopen(argv[1], "rb")))
        fatal("%s: %s\n", argv[1], strerror(errno));

    if (GET16(f) != 0xfefe)
        fatal("%s: not a tokenized MAC/65 file\n", argv[1]);

    fseek(f, 0, SEEK_END);
    flen = ftell(f);
    fseek(f, 2, SEEK_SET);

    if (GET16(f)+4 != flen)
        fatal("%s: invalid length field\n", argv[1]);

    lnum = GET16(f);
    while (!feof(f)) {
//        printf("%5i\t", lnum);      // line number

        len = fgetc(f) - 3;      // tokenized line length
        fread(line, len, 1, f);

        x = 0;
        y = line[x++];
        if (y >= 128) {                 // label
            y -= 128;
            while (y--)
                putc(line[x++], stdout);
            if (x >= len) goto done;
            y = line[x++];
        } else if (!y || y == 88) {     // unparsable line or comment
            while (x<len)
                putc(line[x++], stdout);
            goto done;
        } else if (y == 7) {
            putc('\t', stdout);         // extra tab, macro call
        }

        if (!tok1[y]) fatal("unknown token at %i: %02x\n", ftell(f)-1, y);
        printf("\t%s ", tok1[y]);
    
        for (; x<len;) {
            y = line[x++];
            switch (y) {
            case 5:
                printf("$%04X ", line[x] + (line[x+1]<<8));
                x += 2;
                break;
            case 6:
                printf("$%02X ", line[x++]);
                break;
            case 7:
                printf("%i ", line[x] + (line[x+1]<<8));
                x += 2;
                break;
            case 8:
                printf("%i ", line[x++]);
                break;
            case 10:
                printf("'%c ", line[x++]);
                break;
            case 59:                    // comments until EOL
                putc('\t', stdout);
                while (x<len)
                    putc(line[x++], stdout);
                break;
            default:
                if (y >= 128) {         // label reference
                    y -= 128;
                    while (y--)
                        putc(line[x++], stdout);
                } else {
                    if (!tok2[y]) fatal("unknown token\n");
                    printf("%s", tok2[y]);
                }
            } // switch
        } // for

done:
        printf("\n");
        lnum = GET16(f);
    } // while

    return fclose(f);
}
